package com.insanityengine.ghia.pixels;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.applet.*;

import com.insanityengine.ghia.m3.*;
import com.insanityengine.ghia.libograf.*;

/**
 *
 * <P>
 * This provides a decent implements of most of RendererInterface 
 * with the exception of the <i>drawPolygon</i> method.  It combines
 * a PixelImage with a ZeeBuffer to provide core rendering functionality.
 * </P>
 *
 * @author BrianHammond
 *
 * $Header: /usr/local/cvsroot/ghia/src/java/com/insanityengine/ghia/pixels/BufferBase.java,v 1.3 2005/03/19 17:50:02 brian Exp $
 *
 */

public class BufferBase extends PixelImage {
	
	///////////////////////////////////////////////////

	/**
	 * 
	 * Constructor
	 * 
	 */
	public BufferBase() {}

	/**
	 *
	 * Initialize the BufferBase, the component is not critical
	 *
	 * @param width of the display
	 * @param height of the display
	 * @param component to use to create offscreen buffer
	 *
	 */
	public void init( int width, int height, Component component ) {
		init( width, height );
	}
	
	/**
	 *
	 * Initialize the BufferBase, the component is not critical
	 *
	 * @param width of the display
	 * @param height of the display
	 *
	 */
	public void init( int width, int height ) {
		super.init( width, height );
		zbuffer.init( width, height );
		frustrum.setSize( width, height );
	}
	
	/**
	 *
	 * Draw a polygon
	 *
	 * @param count of points
	 * @param ptz the array of Pt3 to draw
	 *
	 */
	public void drawPolygon( int pt_idx, Pt3 ptz[] ) {
		// noop
	}
	
	/**
	 *
	 * Set the current ImageSkin
	 *
	 * @param skin to use
	 *
	 */
	public final void setSkin( ImageSkin skin ) {
		this.skin = skin;
	}
	
	/**
	 *
	 * Set the current normal 
	 *
	 * @param point to use as the current normal
	 *
	 */
	public final void setNormal( Pt3 n ) { 
		normal.set( n );
	}

	/**
 	 *
 	 * Set the far value for the zbuffer
 	 *
 	 * @param z to use as far value for the ZeeBuffer
 	 *
 	 */
	public final void setZFar( float z ) {
		zbuffer.setFar( z );
	}

	/**
 	 *
 	 * Get the far value for the zbuffer
 	 *
 	 * @return the far value for the zbuffer
 	 *
 	 */
	public final float getZFar() {
		return zbuffer.getFar();
	}
	
	/**
	 *
	 * Set blending policy
	 *
	 * @param blend or not...
	 *
	 */
	public final void setBlending( boolean blend ) {
		blending = blend;
	}
	
	/**
	 *
	 * Set depth testing policy
	 *
	 * @param testDepth false disables test
	 *
	 */
	public final void setDepthTesting( boolean b ) {
		depthtest = b;
	}

	/**
	 *
	 * Set the shading method
	 *
	 * @param method may be listed in M3_Constants
	 *
	 */
	public final void setShadingMethod( int n ) {
		shade_method = n;
		if ( M3_Constants.PHONG_SHADE == n && null == phongo ) {
			initPhongShading( 64 );
		}
	}
	
	/**
	 *
	 * Set the texture function
	 *
	 * @param function may be listed in M3_Constants
	 *
	 */
	public final void setTextureFunction( int funk ) {
		if ( null != skin ) skin.setFunction( funk );
	}

	/**
	 *
	 * Set the pixel at index to color if this is the closes z
	 * value so far
	 *
	 * @param index to try to set
	 * @param z value to try with
	 * @param color to use (ARGB)
	 *
	 */
	public final void setColor( int index, float z, int color ) {
		if ( zbuffer.set( index, z ) ) super.setColor( index, color );
	}

	/**
	 *
	 * Set the pixel at index corresponding to (x,y) to color 
	 * if this is the closes z value so far
	 *
	 * @param x coordinate
	 * @param y coordinate
	 * @param z value to try with
	 * @param color to use (ARGB)
	 *
	 */
	public final void setColor( float x, float y, float z, int color ) {
		setColor( ( int ) ( x + y * w ), color );
	}

	/**
	 *
	 * Set the pixel at index corresponding to (x,y) to color 
	 * if this is the closes z value so far
	 *
	 * @param x coordinate
	 * @param y coordinate
	 * @param z value to try with
	 * @param alpha color component
	 * @param red color component
	 * @param green color component
	 * @param blue color component
	 *
	 */
	public final void setColor( float x, float y, float z, int alpha, int red, int green, int blue ) {
		setColor( 
			x, 
			y, 
			z, 
			alpha << 24 | red << 16 | green << 8 | blue
		);
	}
	
	/**
	 *
	 * Clear the depth buffer
	 *
	 */
	public final void clearDepthBuffer() {
		zbuffer.clear();
	}

	/**
	 *
	 * Clear the pixel/color buffer
	 *
	 */
	public final void clearColorBuffer() {
		super.clear();
	}

	/**
	 *
	 * Clear the buffer
	 *
	 */
	public final void clear() {
		clearDepthBuffer();
		clearColorBuffer();
	}

	////////////////////////////////////////////////////////////////////
	
	private Pt3 tri_lame[] = new Pt3[ 3 ];

	private boolean blending = false;
	
	private boolean depthtest = true;

	// do these need to be private and have accessors?
	
	protected Pt3 normal = new Pt3();
	
	protected int phongo[] = null;
	protected int shade_method = M3_Constants.NONE_SHADE;
	protected int textFunk = M3_Constants.TEXT_WRAP;

	protected ImageSkin skin = null;
	protected ZeeBuffer zbuffer = new ZeeBuffer();
	protected Frustrum frustrum = new Frustrum();
	
	////////////////////////////////////////////////////////////////////

	private final void initPhongShading( int n ) {
		phongo = new int[ 256 ];

		int clrL = ( int ) 232;
		int clrI = ( int ) ( 255 - clrL );

		int phF1 = ( int ) ( 255 - n );
		int phF2 = ( int ) ( n - 255 );

		for ( int i = 0 ; i < phF1 ; i++ ) {
			phongo[ i ] = ( int ) ( clrL * i / phF1 ); 
		}
		for ( int i = phF1 ; i < 256 ; i++ ) {
			phongo[ i ] = ( int ) ( 
				clrL + ( clrI * ( i + phF2 ) / n )
			);
		}
	}
};

/**
 *
 * $Log: BufferBase.java,v $
 * Revision 1.3  2005/03/19 17:50:02  brian
 * repackaging
 *
 * Revision 1.2  2005/03/12 04:58:47  brian
 * call the methods set instead of copy
 *
 * Revision 1.1  2004/09/02 13:17:16  brian
 * the big reorg
 *
 * Revision 1.15  2004/09/01 01:10:42  brian
 * fix class level javadoc placement
 *
 * Revision 1.14  2004/09/01 00:11:06  brian
 * author, log and header stuff
 *
 *
 */
