package com.insanityengine.ghia.renderer;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.applet.*;

import com.insanityengine.ghia.m3.*;
import com.insanityengine.ghia.libograf.*;
import com.insanityengine.ghia.pixels.*;

/**
 *
 * <P>
 * This renderer uses the BufferBase for the bulk of it's functionality
 * but implements the tricky drawPolygon method
 * </P>
 *
 * <P>
 * It uses two buffers the height of the display and uses a simplistic
 * line drawing algorithm to update the maximum and mininum x values 
 * for each scanline as well as information about the depth and texture
 * coordinate at each point. (scanPolyEdge)
 * </P>
 *
 * <P>
 * Obviously this algorithm only works for concave polygons. If you have
 * convex polygons, you are a freak. I'm sorry, but it's true.
 * </P>
 *
 * <P>
 * Once each edge of the polygon have been scanned, each scanline is 
 * rasterized with an embarrasingly bad implementation which uses the 
 * information gathered above, the current skin, shading mode, etc
 * to raster the scanline with zbuffer lovin'.  (rasterPoly)
 * </P>
 *
 * <P>
 * I'm sure your implementation will be much better, all you need to do is
 * derive a class from BufferBase and implement the drawPolygon method or 
 * implement RendererInterface.
 * </P>
 *
 * <table border="1">
 *  <topic>TestResults</topic>
 *	<TH>TestName</TH>
 *	<TH>Results</TH>
 *	<TH>SnapShot</TH>
 *  <TR valign="top">
 *   <TD>Light</TD>
 *   <TD>51fps</TD>
 *   <TD><img src="http://ghia.insanityengine.com/images/old_snaps/Memagery_Light.gif"/></TD>
 *  </TR>
 *  <TR valign="top">
 *   <TD>GeomLoader</TD>
 *   <TD>1102/61.640=17.878002 * 4725 =  84475 tps =  84 kts</TD>
 *   <TD><img src="http://ghia.insanityengine.com/images/old_snaps/Memagery_Geom.gif"/></TD>
 *  </TR>
 * </TABLE>
 *
 * @author BrianHammond
 *
 * $Header: /usr/local/cvsroot/ghia/src/java/com/insanityengine/ghia/renderer/Memagery.java,v 1.9 2005/03/26 23:42:41 brian Exp $
 *
 */
public class Memagery extends RendererBase {

	/**
  	 *	  
  	 * Constructor	  
  	 *	  
  	 */
	public Memagery() {
	}

	/**
	 *
	 * Draw a polygon
	 *
	 * @param count of points
	 * @param ptz the array of Pt3 to draw
	 *
	 */
	public void drawPolygon( int number, Pt3 ptz[] ) {

		if ( normal.z < 0 ) return;
		
		// convert to screen coordinate
		frustrum.toScreen( number, ptz );	
		if ( frustrum.offScreen( number, ptz ) ) return;

		initScantron(); 

		// scan the polygon edges
		float maxz, minz;
		maxz = minz = ptz[ number - 1 ].z;

		int i, j;	
		for ( i = 0 ; i < number ; i++ ) {
			j = i + 1;
			if ( j >= number ) j = 0;
			
			scannie.scanPolyEdge( ptz[ i ], ptz[ j ] );	
			
			if ( ptz[ i ].z > maxz ) maxz = ptz[ i ].z;
			if ( ptz[ i ].z < minz ) minz = ptz[ i ].z;
		}

		// draw the actual scanlines
		rasterPoly( minz, maxz );
	}

	///////////////////////////////////////////////////////////

	/** 
	 * 
	 * rasterPoly: blech
	 * 
	 * @param minz
	 * @param maxz
	 * 
	 */
	private final void rasterPoly( float minz, float maxz ) {
		int i, j, n;
		float diffz = maxz - minz;

		if ( 0 == diffz ) diffz = 1;

		int x0, x1, xd;
		float z, zinc, fz = 1;

		// so colorful!
		int pix =  0xFF00FF00;

		int idx, edx, bidx, pidx;

		// texture and image space coord respectively
		float tx, ty, txi, tyi;
		int isx, isy, iidx;
		
		// pukey!
		int ix, iy;
		int img_w = 0, img_h = 0;

		int shadeMask;

		if ( null != skin ) {
			img_w = skin.getWidth();
			img_h = skin.getHeight();
		}
		
		if ( M3_Constants.NORM_SHADE == shade_method ) fz = 0.1f + 0.9f * normal.z;

		for ( i = 0, bidx = 0 ; i < h ; i++, bidx += w ) {
			if ( !scannie.hasCandy( i ) ) continue;
		
			xd = ( int ) scannie.getDistance( i );
			
			// move in z
			z = scannie.getLz( i );
			zinc = ( scannie.getRz( i ) - scannie.getLz( i ) ) / xd;

			// move in texture space
			tx = ( scannie.getLtx( i )* img_w );
			ty = ( scannie.getLty( i )* img_h );

			txi = ( scannie.getRtx( i ) * img_w - tx ) / xd;
			tyi = ( scannie.getRty( i ) * img_h - ty ) / xd;

			// shading for each pixel on the scanline
			z = scannie.getLz( i );
			j = ( int ) scannie.getLx( i );
			
			idx = bidx + j;
			edx = idx + xd;
			for ( 
				;
				idx <= edx && j < w ; 
				idx++, j++, z += zinc,
				tx += txi, ty += tyi
			) {
				if ( j < 0 ) continue;

				if ( !zbuffer.set( idx, z ) ) continue;
				
				if ( null != skin ) {
					pix = ( int ) skin.getPixelAt( ( int ) ( tx ), ( int ) ( ty ) );
				}

				switch( shade_method ) {
						
					//case NONE_SHADE: 
					default:
						shadeMask = 0xFFFFFFFF;
						break;

					case M3_Constants.NORM_SHADE:
						n = ( int ) ( 255 * fz );
						shadeMask = ( 0xFF << 24 | n << 16 | n << 8 | n );
						break;

					case M3_Constants.DEPTH_SHADE:
						if ( 0 == diffz ) fz = 0.1f;
						else fz = ( z - minz ) / diffz;
						if ( fz < 0.1 ) fz = 0.1f;

						n = ( int ) ( 255 * fz );
						shadeMask = ( 0xFF << 24 | n << 16 | n << 8 | n );
						break;
						
					case M3_Constants.PHONG_SHADE:
						if ( 0 == diffz ) fz = 1;
						else fz = ( z - minz ) / diffz;

						pidx = ( int ) ( 255 * fz );

						if ( pidx < 0 ) pidx = 0;
						if ( pidx > 255 ) pidx = 255;

						n = ( int ) phongo[ pidx ];
						shadeMask = ( 0xFF << 24 | n << 16 | n << 8 | n );
						break;
				}

				// set the pixel
				pixels[ idx ] = pix & shadeMask;
			}
		}
	}

	//////////////////////////////////

	private Scantron scannie;
	private DrawerRegistry registry = new DrawerRegistry();

	/** 
	 * 
	 * initScantron
	 * 
	 */
	final private void initScantron() {
		if ( null == scannie ) {
			scannie = new Scantron( w, h );
		}
		scannie.reset();
	}
};

/**
 *
 * $Log: Memagery.java,v $
 * Revision 1.9  2005/03/26 23:42:41  brian
 * fix links
 *
 * Revision 1.8  2005/03/25 03:07:54  brian
 * centralize more cruft
 *
 * Revision 1.7  2005/03/25 02:45:25  brian
 * split up a bit
 *
 * Revision 1.6  2005/03/19 17:50:02  brian
 * repackaging
 *
 * Revision 1.5  2005/02/17 03:31:13  brian
 * good stuff???
 *
 * Revision 1.4  2004/09/11 19:40:57  brian
 * use normal.z to bail early
 *
 * Revision 1.3  2004/09/11 14:37:21  brian
 * added RendererConsumer and stuff
 *
 * Revision 1.2  2004/09/05 03:41:31  brian
 * refactored to have the render method of the renderers invoke the drawingIfz
 *
 * Revision 1.1  2004/09/02 13:17:16  brian
 * the big reorg
 *
 * Revision 1.10  2004/09/01 01:52:47  brian
 * added test results (ok, and images) to javadoc
 *
 * Revision 1.9  2004/09/01 01:10:42  brian
 * fix class level javadoc placement
 *
 * Revision 1.8  2004/09/01 00:11:06  brian
 * author, log and header stuff
 *
 *
 */
